#ifndef MRK_LEXER
#define MRK_LEXER

#include <stdbool.h>

#include "mrk/common.h"

#define MRK_MAX_HEADER_LEN 6
#define MRK_MIN_HORIZ_RULE_LEN 3

typedef struct mrk_lexer mrk_lexer;

typedef enum mrk_lexer_err {
  mrk_lexer_err_ok = 0,
  mrk_lexer_err_done,
  mrk_lexer_err_unexpected_char,
} mrk_lexer_err;

typedef enum mrk_token_type {
  mrk_token_type_none = 0,
  mrk_token_type_backtick,
  mrk_token_type_triple_backtick,
  mrk_token_type_underscore,
  mrk_token_type_double_underscore,
  mrk_token_type_star,
  mrk_token_type_double_star,
  mrk_token_type_newline,
  mrk_token_type_blank_line,
  mrk_token_type_line_break,
  mrk_token_type_right_angle_brackets,
  mrk_token_type_left_bracket,
  mrk_token_type_right_bracket,
  mrk_token_type_left_paren,
  mrk_token_type_right_paren,
  mrk_token_type_bang,
  mrk_token_type_backslash,
  mrk_token_type_text,
  mrk_token_type_header_start,
  mrk_token_type_horizontal_rule,
  mrk_token_type_indent,
  mrk_token_type_list_item_unordered,
  mrk_token_type_list_item_ordered,
  mrk_token_type_list_item_checked,
  mrk_token_type_list_item_unchecked,
} mrk_token_type;

extern const char *mrk_token_type_names[];

typedef struct mrk_token {
  mrk_token_type type;
  size_t start;
  size_t end;
  size_t start_line;
  size_t start_line_index;
} mrk_token;

/**
 * Return the length of the given token.
 */
size_t mrk_token_len(mrk_token t);

/**
 * Return a pointer to the start of the lexeme in the initial buffer, returning
 * the length of the lexeme.
 */
size_t mrk_token_lexeme(const char **out, mrk_lexer *lexer, mrk_token t);

/**
 * Initialize a new lexer struct.
 */
mrk_err mrk_lexer_init(mrk_lexer **out);

/**
 * Deallocate the lexer object.
 */
void mrk_lexer_free(mrk_lexer *lexer);

/**
 * Open the buffer with the given lexer struct. `buf` is expected to live for
 * the duration of the lexing.
 *
 * The lexer will run either until `len` characters have been matched, or until
 * a nul character has been reached. If `len` is set to 0, only the nul check is
 * used to determine the end of the buffer.
 */
void mrk_lexer_open(mrk_lexer *lexer, const char *buf, size_t len);

/**
 * Returns whether the lexer is done.
 */
bool mrk_lexer_done(const mrk_lexer *lexer);

/**
 * Output the next lexed token for the given input.
 */
mrk_lexer_err mrk_lexer_next(mrk_token *out, mrk_lexer *lexer);

#endif
