#ifndef MRK_LEXER_INTERNAL
#define MRK_LEXER_INTERNAL

#include "mrk/lexer.h"

struct mrk_lexer {
  struct {
    const char *s;
    size_t len;
  } buf;
  struct {
    size_t line;
    size_t line_index;
    size_t buf_index;
  } pos;
  mrk_token cur_token;
  mrk_token last_emitted;
};

/**
 * Return the next character that would be consumed by a call to advance. At the
 * end of the buffer, this value is nul.
 */
char mrk_lexer_peek(mrk_lexer *lexer);

/**
 * Return the n'th next character that would be consumed. If `n` is zero, this
 * function is equivalent to calling peek.
 */
char mrk_lexer_peek_n(mrk_lexer *lexer, size_t n);

/**
 * Returns true if the nul-terminated string s matches the next characters that
 * would be consumed. This is a convenience method instead of having to call
 * multiple peek calls.
 */
bool mrk_lexer_peek_str(mrk_lexer *lexer, const char *s);

/**
 * Advance the current position by one character, adding the new character to
 * the curent token's context and returning it.
 */
char mrk_lexer_advance(mrk_lexer *lexer);

/**
 * Advance until the next element to peek is not equal to c.
 */
void mrk_lexer_advance_eq(mrk_lexer *lexer, char c);

/**
 * Advance `n` positions; equivalent to running advance `n` times and returning
 * the last call's result.
 */
char mrk_lexer_advance_n(mrk_lexer *lexer, size_t n);

/**
 *  Advance the lexer as long as the next character is part of a regular word.
 */
void mrk_lexer_advance_word(mrk_lexer *lexer);

/**
 * Reset the lexer's current token context.
 */
void mrk_lexer_reset(mrk_lexer *lexer);

/**
 * Output the currently matched token to the token struct with the given type,
 * and reset the lexer's tracked token.
 */
void mrk_lexer_emit(mrk_token *out, mrk_lexer *lexer, mrk_token_type type);

/**
 * Returns whether the given character can be part of a word, and is thus not a
 * special character.
 */
bool mrk_is_special_char(char c);

/**
 * Lex a token at the start of a line (including after 1 or more indents).
 */
void mrk_lexer_lex_start_of_line(mrk_token *out, mrk_lexer *lexer);

/**
 * Lex a token that starts in the middle of a line.
 */
void mrk_lexer_lex_middle_of_line(mrk_token *out, mrk_lexer *lexer);

/**
 * Advance the lexer until it encounters a character that's always parsed as a
 * separate token, regardless of context (e,g. newlines, brackets).
 */
void mrk_lexer_advance_text(mrk_lexer *lexer);

#endif
