#include "mrk/parser.h"
#include "mrk/lexer.h"
#include "mrk/parser_internal.h"
#include <stdio.h>

mrk_err mrk_parser_init(mrk_parser **out) {
  MRK_CALLOC(out, 1, sizeof(mrk_parser));

  return mrk_err_ok;
}

void mrk_parser_open(mrk_parser *parser, mrk_lexer *lexer) {
  parser->lexer = lexer;

  // Prefill lookahead character
  mrk_parser_advance(parser);
}

mrk_err mrk_parser_parse(mrk_ast_node **out, mrk_parser *parser) {
  mrk_ast_node *root;
  MRK_RES(mrk_ast_node_init(&root));
  mrk_parser_indent(parser);

  while (!mrk_parser_done(parser)) {
    MRK_RES(mrk_parser_parse_indent_block(parser, root));
  }

  *out = root;

  return mrk_err_ok;
}

mrk_err mrk_parser_parse_indent_block(mrk_parser *parser,
                                      mrk_ast_node *parent) {
  size_t indent = parser->indent;

  while (!mrk_parser_done(parser) && parser->indent == indent) {
    mrk_ast_node *child;
    MRK_RES(mrk_ast_node_child_append(&child, parent));

    switch (mrk_parser_peek(parser).type) {
    case mrk_token_type_header_start:
      MRK_RES(mrk_parser_parse_header(child, parser));
      break;
    case mrk_token_type_text:
    case mrk_token_type_left_bracket:
      MRK_RES(mrk_parser_parse_paragraph(child, parser));
      break;
    case mrk_token_type_list_item_unordered:
    case mrk_token_type_list_item_checked:
    case mrk_token_type_list_item_unchecked:
    case mrk_token_type_list_item_ordered:
      MRK_RES(mrk_parser_parse_list(child, parser));
      break;
    default:
      parser->error.code = mrk_parser_err_unexpected_token;
      parser->error.token = mrk_parser_peek(parser);
      return mrk_err_invalid_md;
    }
  }

  return mrk_err_ok;
}

mrk_err mrk_parser_parse_header(mrk_ast_node *out, mrk_parser *parser) {
  mrk_token header_token;
  mrk_parser_eat(&header_token, parser, mrk_token_type_header_start);

  out->type = mrk_ast_node_type_header;
  out->d.header.depth = mrk_token_len(header_token);

  // All continuation lines of the header should be one indentation deeper than
  // the definition of the header
  parser->indent++;

  MRK_RES(mrk_parser_parse_common(parser, out));

  return mrk_err_ok;
}

mrk_err mrk_parser_parse_text(mrk_ast_node *out, mrk_parser *parser) {
  mrk_token text_token;
  MRK_RES(mrk_parser_eat(&text_token, parser, mrk_token_type_text));

  out->type = mrk_ast_node_type_text;
  out->d.text.start = text_token.start;
  out->d.text.end = text_token.end;

  return mrk_err_ok;
}

mrk_err mrk_parser_parse_common(mrk_parser *parser, mrk_ast_node *parent) {
  size_t indent = parser->indent;

  while (!mrk_parser_done(parser) && parser->indent == indent) {
    mrk_ast_node *child;

    mrk_token token = mrk_parser_peek(parser);

    switch (token.type) {
    case mrk_token_type_text:
      MRK_RES(mrk_ast_node_child_append(&child, parent));
      MRK_RES(mrk_parser_parse_text(child, parser));
      break;
    case mrk_token_type_left_bracket:
      MRK_RES(mrk_ast_node_child_append(&child, parent));
      MRK_RES(mrk_parser_parse_link(child, parser));
      break;
    case mrk_token_type_backtick:

      MRK_RES(mrk_ast_node_child_append(&child, parent));
      MRK_RES(mrk_parser_parse_code(child, parser));
      break;
    case mrk_token_type_newline:
      MRK_RES(mrk_ast_node_child_append(&child, parent));
      child->type = mrk_ast_node_type_space;
      mrk_parser_advance(parser);
      mrk_parser_indent(parser);
      break;
    case mrk_token_type_blank_line:
      mrk_parser_advance(parser);
      mrk_parser_indent(parser);
      return mrk_err_ok;
    // All these tokens have no special meaning in this context
    case mrk_token_type_left_paren:
    case mrk_token_type_right_paren:
      mrk_parser_advance(parser);

      MRK_RES(mrk_ast_node_child_append(&child, parent));
      child->type = mrk_ast_node_type_text;
      child->d.text.start = token.start;
      child->d.text.end = token.end;
      break;
    // Any other tokens aren't part of the common section so we just exit
    default:
      return mrk_err_ok;
    }
  }

  return mrk_err_ok;
}

mrk_err mrk_parser_parse_link(mrk_ast_node *out, mrk_parser *parser) {
  out->type = mrk_ast_node_type_link;

  mrk_token left_bracket;
  MRK_RES(mrk_parser_eat(&left_bracket, parser, mrk_token_type_left_bracket));

  if (mrk_parser_done(parser)) {
    parser->error.code = mrk_parser_err_unclosed_brackets;
    parser->error.token = left_bracket;

    return mrk_err_invalid_md;
  }

  mrk_ast_node *child;
  MRK_RES(mrk_ast_node_child_append(&child, out));

  switch (mrk_parser_peek(parser).type) {
  case mrk_token_type_text:
    MRK_RES(mrk_parser_parse_text(child, parser));
    break;
  // TODO allow other types of text (e.g. cursive)
  // TODO image links
  default:
    // TODO throw error
    break;
  }

  MRK_RES(mrk_parser_eat(NULL, parser, mrk_token_type_right_bracket));
  MRK_RES(mrk_parser_eat(NULL, parser, mrk_token_type_left_paren));

  mrk_token url_token;
  MRK_RES(mrk_parser_eat(&url_token, parser, mrk_token_type_text));

  MRK_RES(mrk_parser_eat(NULL, parser, mrk_token_type_right_paren));

  out->d.link.url_start = url_token.start;
  out->d.link.url_end = url_token.end;

  return mrk_err_ok;
}

mrk_err mrk_parser_parse_paragraph(mrk_ast_node *out, mrk_parser *parser) {
  out->type = mrk_ast_node_type_paragraph;

  MRK_RES(mrk_parser_parse_common(parser, out));

  return mrk_err_ok;
}

mrk_err mrk_parser_parse_list(mrk_ast_node *out, mrk_parser *parser) {
  out->type = mrk_ast_node_type_list;

  size_t indent = parser->indent;

  switch (mrk_parser_peek(parser).type) {
  case mrk_token_type_list_item_unordered:
  case mrk_token_type_list_item_checked:
  case mrk_token_type_list_item_unchecked:
    out->d.list.ordered = false;
    break;
  case mrk_token_type_list_item_ordered:
    out->d.list.ordered = true;
    break;
  default:
    parser->error.code = mrk_parser_err_unexpected_path;
    return mrk_err_invalid_md;
  }

  mrk_ast_node *child;
  MRK_RES(mrk_ast_node_child_append(&child, out));
  MRK_RES(mrk_parser_parse_list_item(child, parser));

  while (!mrk_parser_done(parser) && parser->indent == indent) {
    switch (mrk_parser_peek(parser).type) {
    case mrk_token_type_list_item_unordered:
    case mrk_token_type_list_item_checked:
    case mrk_token_type_list_item_unchecked:
      // Two lists of different types can follow up on each other
      if (out->d.list.ordered) {
        return mrk_err_ok;
      }

      MRK_RES(mrk_ast_node_child_append(&child, out));
      MRK_RES(mrk_parser_parse_list_item(child, parser));
      break;
    case mrk_token_type_list_item_ordered:
      // Two lists of different types can follow up on each other
      if (!out->d.list.ordered) {
        return mrk_err_ok;
      }

      MRK_RES(mrk_ast_node_child_append(&child, out));
      MRK_RES(mrk_parser_parse_list_item(child, parser));
      break;
    // Other tokens imply end of list
    default:
      return mrk_err_ok;
    }
  }

  return mrk_err_ok;
}

mrk_err mrk_parser_parse_list_item(mrk_ast_node *out, mrk_parser *parser) {
  out->type = mrk_ast_node_type_list_item;

  size_t indent = parser->indent;

  switch (mrk_parser_peek(parser).type) {
  case mrk_token_type_list_item_unordered:
  case mrk_token_type_list_item_ordered:
    out->d.list_item.checkbox = false;
    out->d.list_item.checked = false;
    break;
  case mrk_token_type_list_item_checked:
    out->d.list_item.checkbox = true;
    out->d.list_item.checked = true;
    break;
  case mrk_token_type_list_item_unchecked:
    out->d.list_item.checkbox = true;
    out->d.list_item.checked = false;
    break;
  // This path should never be taken
  default:
    parser->error.code = mrk_parser_err_unexpected_path;
    return mrk_err_invalid_md;
  }

  mrk_parser_advance(parser);

  // Parse the textual part of the list
  parser->indent++;
  MRK_RES(mrk_parser_parse_common(parser, out));

  // This allows lists to be arbitrarily nested
  if (!mrk_parser_done(parser) && parser->indent > indent) {
    MRK_RES(mrk_parser_parse_indent_block(parser, out));
  }

  return mrk_err_ok;
}

mrk_err mrk_parser_parse_code(mrk_ast_node *out, mrk_parser *parser) {
  out->type = mrk_ast_node_type_code;

  mrk_token start_backtick;
  MRK_RES(mrk_parser_eat(&start_backtick, parser, mrk_token_type_backtick));

  size_t indent = parser->indent;

  while (!mrk_parser_done(parser) && parser->indent == indent) {
    mrk_ast_node *child;
    mrk_token token = mrk_parser_peek(parser);

    switch (token.type) {
    case mrk_token_type_blank_line:
      parser->error.code = mrk_parser_err_unclosed_brackets;
      parser->error.token = start_backtick;

      return mrk_err_invalid_md;
    case mrk_token_type_backtick:
      mrk_parser_advance(parser);

      return mrk_err_ok;
    case mrk_token_type_newline:
      MRK_RES(mrk_ast_node_child_append(&child, out));
      child->type = mrk_ast_node_type_space;
      mrk_parser_advance(parser);
      mrk_parser_indent(parser);
      break;
    default:
      mrk_parser_advance(parser);

      if (out->children.len == 0 ||
          out->children.arr[out->children.len - 1]->type !=
              mrk_ast_node_type_text) {
        MRK_RES(mrk_ast_node_child_append(&child, out));
        child->type = mrk_ast_node_type_text;
        child->d.text.start = token.start;
        child->d.text.end = token.end;
      }
      // Simply append current text to previous one
      else {
        child = out->children.arr[out->children.len - 1];
        child->d.text.end = token.end;
      }
      break;
    }
  }

  // Successful exit of the function should've already happened in the while
  // loop
  parser->error.code = mrk_parser_err_unclosed_brackets;
  parser->error.token = start_backtick;

  return mrk_err_invalid_md;
}
